using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using System.Diagnostics;

namespace Stopper {
	public class ClothSimProxy : MVRScript {
		private string fileExt = "simproxywrap";

		bool selecting = false;
		bool selectingClothing = false;
		bool invertedMode = false;

        DAZDynamic dazDynamic;

		DAZClothingItem clothingItem;
        DAZClothingItem proxyClothingItem;
        
        DAZSkinWrap originalPrimarySkinWrap;
		DAZSkinWrap originalProxySkinWrap;

        SkinWrapReplacer simProxyWrap;
        SkinWrapReplacer primaryWrap;

        DummyDAZSkinV2 dummySkin;

        DAZCharacterSelector selector;

        MVRScriptExtensions.VerticalLayoutPanel mainPanel;
        MVRScriptExtensions.VerticalLayoutPanel selectionPanel;

		DAZSkinWrapStore wrapStore;

		JSONStorableStringChooser simProxyChooser;
		UIDynamicPopup simProxyChooserPopup;

		JSONStorableUrl wrapStoreFilePath;

        public class DummyDAZSkinV2 : DAZSkinV2 {
            private SkinWrapReplacer skinWrap;
            private List<ComputeBuffer> createdBuffers = new List<ComputeBuffer>();

            public DummyDAZSkinV2() {
                this.GPUAutoSwapShader = false;
				this.delayDisplayOneFrame = false;
            }

            public void InitializeMesh() {
                this.InitMesh();
            }

            public void ConnectSkinWrap(SkinWrapReplacer skinWrap) {
                this.skinWrap = skinWrap;
                this.dazMesh = skinWrap.dazMesh;
                this.CopyMaterials();
                this.InitializeMesh();
                
                this.rawVertsBuffer = skinWrap.verticesBuffer1;
                this.surfaceNormalsBufferAccessor = skinWrap.surfaceNormalsBuffer;

                // Someone might reload the proxy cloth, which will release the buffers
                skinWrap.preDestroyHandlers += () => {
                    this.skinWrap = null;          
                    this.dazMesh = null;          
                    this.rawVertsBuffer = CopyBuffer(skinWrap.drawVerticesBuffer);
                    this.surfaceNormalsBufferAccessor = CopyBuffer(skinWrap.surfaceNormalsBuffer);
                };
            }

            private ComputeBuffer CopyBuffer(ComputeBuffer buffer) {
                byte[] data = new byte[buffer.count * buffer.stride];
                buffer.GetData(data);

                ComputeBuffer copy = new ComputeBuffer(buffer.count, buffer.stride);
                createdBuffers.Add(copy);
                copy.SetData(data);
                return copy;
            }

            public ComputeBuffer surfaceNormalsBufferAccessor {
                get {
                    return _surfaceNormalsBuffer;
                } set {
                    _surfaceNormalsBuffer = value;
                }
            }

            public void LateUpdate() {
                if(skinWrap != null && skinWrap.drawVerticesBuffer != null && this.rawVertsBuffer != skinWrap.drawVerticesBuffer) {
                    this.rawVertsBuffer = skinWrap.drawVerticesBuffer;
                }
            }

            private void OnDestroy() {
                foreach(var buffer in createdBuffers) {
                    buffer.Release();
                }
                createdBuffers.Clear();
            }
        }
		
		public override void Init() {
            WalkAndGetParentItems();

            if(clothingItem == null) {
                throw new Exception("Unable to find containing clothing item");
            }

            selector = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;

			JSONStorableStringChooser simProxyChooser = new JSONStorableStringChooser(
                        "Sim Proxy Uid", 
                        selector.clothingItems.Select(c => c.uid).ToList(),
                        null, 
                        "Sim Proxy Uid", 
                        (string uid) => {
							proxyClothingItem = selector.GetClothingItem(uid);
							StartCoroutine(ApplySimProxy(proxyClothingItem));
                        });
			RegisterStringChooser(simProxyChooser);
			simProxyChooserPopup = CreateFilterablePopup(simProxyChooser, false);
			simProxyChooserPopup.popup.onOpenPopupHandlers += () => {
				simProxyChooser.choices = selector.clothingItems.Select(c => c.uid).ToList();
			};

			wrapStoreFilePath = new JSONStorableUrl("Loaded Wrap", "");
			RegisterUrl(wrapStoreFilePath);
        }

        private void WalkAndGetParentItems(){
			GameObject parent = this.gameObject;
			while(parent != null && parent != containingAtom.gameObject){
				if(parent.transform.parent == null){
					return;
				}
				
				if(dazDynamic == null){
					dazDynamic = parent.GetComponent<DAZDynamic>();
				}

				if(clothingItem == null){
					clothingItem = parent.GetComponent<DAZClothingItem>();
				}

				parent = parent.transform.parent.gameObject;
			}
		}

        public void InitMainUI() {
            if(mainPanel == null) {


                var mainUI = this.rightUIContent.transform.parent;
                mainPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("MainPanel", mainUI);

                selectionPanel = this.AddUIElement<MVRScriptExtensions.VerticalLayoutPanel>("SelectionPanel", mainPanel);
                
                this.AddToPanel(simProxyChooserPopup, selectionPanel);
       
                var wrapButton = CreateButton("Re-Apply Sim Cloth And Wrap");
                wrapButton.button.onClick.AddListener(() => {
                    if(proxyClothingItem != null) {
                        StartCoroutine(ApplySimProxy(proxyClothingItem));
                    }
                });
                this.AddToPanel(wrapButton, selectionPanel);
                wrapButton.height = 70;

				var savePath = dazDynamic.GetStoreFolderPath(false);
				var saveWrapButton = CreateButton("Save Wrap");
				saveWrapButton.button.onClick.AddListener(() =>
				{
					FileManagerSecure.CreateDirectory(savePath);
					var fileBrowserUI = SuperController.singleton.fileBrowserUI;
					fileBrowserUI.SetTitle("Save Wrap");
					fileBrowserUI.fileRemovePrefix = null;
					fileBrowserUI.hideExtension = false;
					fileBrowserUI.keepOpen = false;
					fileBrowserUI.fileFormat = "simproxywrap";
					fileBrowserUI.defaultPath = savePath;
					fileBrowserUI.showDirs = true;
					fileBrowserUI.shortCuts = null;
					fileBrowserUI.browseVarFilesAsDirectories = false;
					fileBrowserUI.SetTextEntry(true);
					fileBrowserUI.Show(SaveWrap);
					fileBrowserUI.fileEntryField.text = GetFileNameForWrap() + "." + fileExt;
					fileBrowserUI.ActivateFileNameField();
				});
				this.AddToPanel(saveWrapButton, selectionPanel);


				var loadWrapButton = CreateButton("Load Wrap");
				loadWrapButton.button.onClick.AddListener(() =>
				{
					FileManagerSecure.CreateDirectory(savePath);
					var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
					SuperController.singleton.GetMediaPathDialog(LoadWrap, fileExt, savePath, false, true, false, null, false, shortcuts);
				});
				this.AddToPanel(loadWrapButton, selectionPanel);

				var loadedPathLabel = this.AddTextBox($"Loaded Wrap: {wrapStoreFilePath.val}", selectionPanel);
				wrapStoreFilePath.setCallbackFunction += (val) => {
					loadedPathLabel.text = $"Loaded Wrap: {val}";
				};

            }
        }

        public void Start() {
            InitMainUI();
        }

        private IEnumerator ApplySimProxy(DAZClothingItem proxyClothingItem) {
            if(proxyClothingItem == null) {
                yield break;
            }
			
			selector.SetActiveClothingItem(proxyClothingItem, true);

			while(SuperController.singleton.isLoading) {
				yield return null;
			}

			while(!proxyClothingItem.ready || !proxyClothingItem.enabled) {
				yield return null;
			}

			while(!clothingItem.ready || !clothingItem.enabled) {
				yield return null;
			}

			// Wait one more frame
			yield return null;

			// Toggle enable/disable to change the execution order of ClothSettings
			// Otherwise, the primary cloth might render before the sim and be delayed 1 frame
			selector.SetActiveClothingItem(clothingItem, false);
			selector.SetActiveClothingItem(clothingItem, true);

			foreach(DAZSkinWrap w in clothingItem.GetComponentsInChildren<DAZSkinWrap>(true)) {
				bool test = $"{w.GetType()}" == "DAZSkinWrap";
				if($"{w.GetType()}" == "DAZSkinWrap") {
					originalPrimarySkinWrap = w;		
				}
			}

			foreach(DAZSkinWrap w in proxyClothingItem.GetComponentsInChildren<DAZSkinWrap>(true)) {
				if($"{w.GetType()}" == "DAZSkinWrap") {
					originalProxySkinWrap = w;		
				}
			}

			if(originalPrimarySkinWrap == null) {
				SuperController.LogError("Could not find skin wrap");
				yield break;
			}
			
			this.InitSkinWrapSwitcher(originalPrimarySkinWrap);
			this.InitSkinWrapSwitcher(originalProxySkinWrap);

            simProxyWrap = this.ReplaceDazSkinWrap<SkinWrapReplacer>(originalProxySkinWrap); 

            // Don't need this, but this calls some internal initializers
            var normals = simProxyWrap.NormalsBuffer;
            
            primaryWrap = this.ReplaceDazSkinWrap<SkinWrapReplacer>(originalPrimarySkinWrap);

            dummySkin = this.gameObject.AddComponent<DummyDAZSkinV2>();
            dummySkin.ConnectSkinWrap(simProxyWrap);
            primaryWrap.skin = dummySkin;
            primaryWrap.moveToSurface = false;
			simProxyWrap.moveToSurface = false;

			if(wrapStoreFilePath.val != null && FileManagerSecure.FileExists(wrapStoreFilePath.val)) {
				LoadWrap(wrapStoreFilePath.val);
			} else {
				ReWrap();
			}

			originalPrimarySkinWrap.draw = false;
			originalPrimarySkinWrap.enabled = false;
			originalPrimarySkinWrap.renderSuspend = true;

			foreach (ClothSettings clothSettings in simProxyWrap.GetComponents<ClothSettings>())
            {
                //clothSettings.MeshProvider.PreCalcProvider = newSkinWrap;
                clothSettings.builder.physics.DispatchCopyToOld();
                
				clothSettings.MeshProvider.Dispatch();
                clothSettings.builder.physics.Dispatch();
            }
        }

		private void Update() {
			// There's a race condition on startup that somehow results in this desynchronizing 
			if(primaryWrap != null && dummySkin != null && primaryWrap.skin != dummySkin) {
				primaryWrap.skin = dummySkin;
			}
		}

        private void ReWrap(){
			try{
				/* 
					This code was taken from decompiled source of the DAZSkinWrap WrapThreaded function.
				*/
				wrapStore = ScriptableObject.CreateInstance<DAZSkinWrapStore>();
				
				if (primaryWrap.dazMesh != null && primaryWrap.skin != null && primaryWrap.skin.dazMesh != null)
				{
					wrapStore.wrapVertices = new DAZSkinWrapStore.SkinWrapVert[primaryWrap.dazMesh.numUVVertices];
					Vector3[] morphedBaseVertices = primaryWrap.dazMesh.morphedBaseVertices;
					Vector3[] morphedUVNormals = primaryWrap.dazMesh.morphedUVNormals;
					Vector3[] baseSurfaceNormals = primaryWrap.skin.dazMesh.baseSurfaceNormals;
					int[] baseTriangles = primaryWrap.skin.dazMesh.baseTriangles;
					Vector3[] array;
                    /*
					if (skinWrap.wrapToSkinnedVertices)
					{
						if (skinWrap.skin.skinMethod == DAZSkinV2.SkinMethod.CPUAndGPU)
						{
							array = skinWrap.skin.rawSkinnedVerts;
						}
						else
						{
							array = skinWrap.skin.drawVerts;
						}
					}
					else if (skinWrap.wrapToMorphedVertices)
					{
						array = skinWrap.skin.dazMesh.morphedUVVertices;
					}
					else
					{
						array = skinWrap.skin.dazMesh.baseVertices;
					}*/
                    array = primaryWrap.skin.dazMesh.baseVertices;
					for (int i = 0; i < primaryWrap.dazMesh.numUVVertices; i++)
					{
						wrapStore.wrapVertices[i] = default(DAZSkinWrapStore.SkinWrapVert);
					}
					Dictionary<int, bool> dictionary = new Dictionary<int, bool>();
					int[][] baseMaterialVertices = primaryWrap.skin.dazMesh.baseMaterialVertices;
					bool[] array2 = primaryWrap.skin.dazMesh.materialsEnabled;
					for (int j = 0; j < array2.Length; j++)
					{
						if (array2[j])
						{
							for (int k = 0; k < baseMaterialVertices[j].Length; k++)
							{
								if (!dictionary.ContainsKey(baseMaterialVertices[j][k]))
								{
									dictionary.Add(baseMaterialVertices[j][k], true);
								}
							}
						}
					}
					for (int l = 0; l < primaryWrap.dazMesh.numBaseVertices; l++)
					{
						
						float num = (float)l * 1f / (float)primaryWrap.dazMesh.numBaseVertices;
						
						int num2 = -1;
						float num3 = 10000f;
						int num4 = -1;
						
						for (int m = 0; m < baseTriangles.Length; m += 3) {
							num4++;
						
							int num5 = baseTriangles[m];
							int num6 = baseTriangles[m + 1];
							int num7 = baseTriangles[m + 2];
							
							// We don't care if the vertex normals are off or the material is disabled
							if ((primaryWrap.vertexIndexLimit == -1 || (num5 < primaryWrap.vertexIndexLimit && num6 < primaryWrap.vertexIndexLimit && num7 < primaryWrap.vertexIndexLimit)) && (!primaryWrap.wrapCheckNormals || Vector3.Dot(baseSurfaceNormals[num4], morphedUVNormals[l]) >= 0f) && (primaryWrap.wrapToDisabledMaterials || (dictionary.ContainsKey(num5) && dictionary.ContainsKey(num6) && dictionary.ContainsKey(num7))))
							{
								float magnitude = ((array[num5] + array[num6] + array[num7]) * 0.33333f - morphedBaseVertices[l]).magnitude;
								if (magnitude < num3)
								{
									num2 = num4;
									num3 = magnitude;
								}
							}
						}
						if (num2 == -1)
						{
							SuperController.LogMessage("Could not find closest triangle during skin wrap");
						}
						DAZSkinWrapStore.SkinWrapVert skinWrapVert = wrapStore.wrapVertices[l];
						skinWrapVert.closestTriangle = num2;
						int num8 = num2 * 3;
						int num9 = baseTriangles[num8];
						int num10 = baseTriangles[num8 + 1];
						int num11 = baseTriangles[num8 + 2];
						float magnitude2 = (array[num9] - morphedBaseVertices[l]).magnitude;
						float magnitude3 = (array[num10] - morphedBaseVertices[l]).magnitude;
						float magnitude4 = (array[num11] - morphedBaseVertices[l]).magnitude;
						if (magnitude2 < magnitude3)
						{
							if (magnitude2 < magnitude4)
							{
								skinWrapVert.Vertex1 = num9;
								if (magnitude3 < magnitude4)
								{
									skinWrapVert.Vertex2 = num10;
									skinWrapVert.Vertex3 = num11;
								}
								else
								{
									skinWrapVert.Vertex2 = num11;
									skinWrapVert.Vertex3 = num10;
								}
							}
							else
							{
								skinWrapVert.Vertex1 = num11;
								skinWrapVert.Vertex2 = num9;
								skinWrapVert.Vertex3 = num10;
							}
						}
						else if (magnitude3 < magnitude4)
						{
							skinWrapVert.Vertex1 = num10;
							if (magnitude2 < magnitude4)
							{
								skinWrapVert.Vertex2 = num9;
								skinWrapVert.Vertex3 = num11;
							}
							else
							{
								skinWrapVert.Vertex2 = num11;
								skinWrapVert.Vertex3 = num9;
							}
						}
						else
						{
							skinWrapVert.Vertex1 = num11;
							skinWrapVert.Vertex2 = num10;
							skinWrapVert.Vertex3 = num9;
						}
						Vector3 vector = (array[skinWrapVert.Vertex1] + array[skinWrapVert.Vertex2] + array[skinWrapVert.Vertex3]) * 0.33333f - array[skinWrapVert.Vertex1];
						Vector3 rhs = baseSurfaceNormals[num2];
						Vector3 rhs2 = Vector3.Cross(vector, rhs);
						Vector3 lhs = morphedBaseVertices[l] - array[skinWrapVert.Vertex1];
						skinWrapVert.surfaceNormalProjection = Vector3.Dot(lhs, rhs);
						skinWrapVert.surfaceTangent1Projection = Vector3.Dot(lhs, vector) / vector.sqrMagnitude;
						skinWrapVert.surfaceTangent2Projection = Vector3.Dot(lhs, rhs2) / rhs2.sqrMagnitude;
						skinWrapVert.surfaceNormalWrapNormalDot = Vector3.Dot(morphedUVNormals[l], rhs);
						skinWrapVert.surfaceTangent1WrapNormalDot = Vector3.Dot(morphedUVNormals[l], vector) / vector.sqrMagnitude;
						skinWrapVert.surfaceTangent2WrapNormalDot = Vector3.Dot(morphedUVNormals[l], rhs2) / rhs2.sqrMagnitude;
						
						wrapStore.wrapVertices[l] = skinWrapVert;
					}
					foreach (DAZVertexMap dazvertexMap in primaryWrap.dazMesh.baseVerticesToUVVertices)
					{
						wrapStore.wrapVertices[dazvertexMap.tovert].closestTriangle = wrapStore.wrapVertices[dazvertexMap.fromvert].closestTriangle;
						wrapStore.wrapVertices[dazvertexMap.tovert].Vertex1 = wrapStore.wrapVertices[dazvertexMap.fromvert].Vertex1;
						wrapStore.wrapVertices[dazvertexMap.tovert].Vertex2 = wrapStore.wrapVertices[dazvertexMap.fromvert].Vertex2;
						wrapStore.wrapVertices[dazvertexMap.tovert].Vertex3 = wrapStore.wrapVertices[dazvertexMap.fromvert].Vertex3;
						wrapStore.wrapVertices[dazvertexMap.tovert].surfaceNormalProjection = wrapStore.wrapVertices[dazvertexMap.fromvert].surfaceNormalProjection;
						wrapStore.wrapVertices[dazvertexMap.tovert].surfaceTangent1Projection = wrapStore.wrapVertices[dazvertexMap.fromvert].surfaceTangent1Projection;
						wrapStore.wrapVertices[dazvertexMap.tovert].surfaceTangent2Projection = wrapStore.wrapVertices[dazvertexMap.fromvert].surfaceTangent2Projection;
						wrapStore.wrapVertices[dazvertexMap.tovert].surfaceNormalWrapNormalDot = wrapStore.wrapVertices[dazvertexMap.fromvert].surfaceNormalWrapNormalDot;
						wrapStore.wrapVertices[dazvertexMap.tovert].surfaceTangent1WrapNormalDot = wrapStore.wrapVertices[dazvertexMap.fromvert].surfaceTangent1WrapNormalDot;
						wrapStore.wrapVertices[dazvertexMap.tovert].surfaceTangent2WrapNormalDot = wrapStore.wrapVertices[dazvertexMap.fromvert].surfaceTangent2WrapNormalDot;
					}
				}

				SetPrimaryWrapStore();
                
			} catch (Exception e){
			}
		}

		private void SetPrimaryWrapStore() {
			primaryWrap.wrapStore = wrapStore;

			primaryWrap.UpdateWrapBuffer();
			primaryWrap.FlushBuffers();
		}

		private string GetFileNameForWrap() {
			return $"{SanitizeFileName(clothingItem.displayName)}_to_{SanitizeFileName(proxyClothingItem.displayName)}";
		}

		// From the dotnet runtime, since plugins don't have access to System.IO.Path
		private static char[] GetInvalidFileNameChars() => new char[] {
			'\"', '<', '>', '|', '\0',
			(char)1, (char)2, (char)3, (char)4, (char)5, (char)6, (char)7, (char)8, (char)9, (char)10,
			(char)11, (char)12, (char)13, (char)14, (char)15, (char)16, (char)17, (char)18, (char)19, (char)20,
			(char)21, (char)22, (char)23, (char)24, (char)25, (char)26, (char)27, (char)28, (char)29, (char)30,
			(char)31, ':', '*', '?', '\\', '/'
		};

		private string SanitizeFileName(string filename){
			return String.Join("_", filename.Split(GetInvalidFileNameChars(), StringSplitOptions.RemoveEmptyEntries) ).TrimEnd('.');
		}

		private void SaveWrap(string path) {
			BinaryStreamWriter binWriter = new BinaryStreamWriter();
			binWriter.Write(wrapStore.wrapVertices.Length);
			for (int i = 0; i < wrapStore.wrapVertices.Length; i++)
			{
				binWriter.Write(wrapStore.wrapVertices[i].closestTriangle);
				binWriter.Write(wrapStore.wrapVertices[i].Vertex1);
				binWriter.Write(wrapStore.wrapVertices[i].Vertex2);
				binWriter.Write(wrapStore.wrapVertices[i].Vertex3);
				binWriter.Write(wrapStore.wrapVertices[i].surfaceNormalProjection);
				binWriter.Write(wrapStore.wrapVertices[i].surfaceTangent1Projection);
				binWriter.Write(wrapStore.wrapVertices[i].surfaceTangent2Projection);
				binWriter.Write(wrapStore.wrapVertices[i].surfaceNormalWrapNormalDot);
				binWriter.Write(wrapStore.wrapVertices[i].surfaceTangent1WrapNormalDot);
				binWriter.Write(wrapStore.wrapVertices[i].surfaceTangent2WrapNormalDot);
			}
			FileManagerSecure.WriteAllBytes(path, binWriter.GetBytes());

			wrapStoreFilePath.val = path;	
		}

		private void LoadWrap(string path) {
			if(primaryWrap == null) {
				SuperController.LogError("Cannot load wrap. Primary wrap not connected.");
				return;
			}

			if(simProxyWrap == null) {
				SuperController.LogError("Cannot load wrap. Primary wrap not connected.");
				return;
			}

			wrapStore = ScriptableObject.CreateInstance<DAZSkinWrapStore>();

			byte[] bytes = FileManagerSecure.ReadAllBytes(path);

			BinaryStreamReader binReader = new BinaryStreamReader(bytes);
			int vertexCount = binReader.ReadInt32();

			if(primaryWrap.dazMesh.numUVVertices != vertexCount) {
				SuperController.LogError($"Cannot apply wrap. Vertex count of wrap ({vertexCount}) does not match that of clothing ({primaryWrap.dazMesh.numUVVertices})");
				return;
			}

			wrapStore.wrapVertices = new DAZSkinWrapStore.SkinWrapVert[vertexCount];
			for (int i = 0; i < vertexCount; i++)
			{
				wrapStore.wrapVertices[i] = default(DAZSkinWrapStore.SkinWrapVert);
				wrapStore.wrapVertices[i].closestTriangle = binReader.ReadInt32();
				wrapStore.wrapVertices[i].Vertex1 = binReader.ReadInt32();
				wrapStore.wrapVertices[i].Vertex2 = binReader.ReadInt32();
				wrapStore.wrapVertices[i].Vertex3 = binReader.ReadInt32();
				wrapStore.wrapVertices[i].surfaceNormalProjection = binReader.ReadSingle();
				wrapStore.wrapVertices[i].surfaceTangent1Projection = binReader.ReadSingle();
				wrapStore.wrapVertices[i].surfaceTangent2Projection = binReader.ReadSingle();
				wrapStore.wrapVertices[i].surfaceNormalWrapNormalDot = binReader.ReadSingle();
				wrapStore.wrapVertices[i].surfaceTangent1WrapNormalDot = binReader.ReadSingle();
				wrapStore.wrapVertices[i].surfaceTangent2WrapNormalDot = binReader.ReadSingle();
			}

			SetPrimaryWrapStore();
			wrapStoreFilePath.val = path;
		}

        private void OnEnable() {
			if(originalPrimarySkinWrap != null && primaryWrap != null) {
				originalPrimarySkinWrap.enabled = false;
				originalPrimarySkinWrap.renderSuspend = true;
				originalPrimarySkinWrap.draw = false;

				primaryWrap.renderSuspend = false;
				primaryWrap.enabled = true;

				this.SwapSkinWrap(originalPrimarySkinWrap, primaryWrap);
				primaryWrap.skin = dummySkin;
			}
			if(simProxyWrap != null) {
				this.SwapSkinWrap(originalProxySkinWrap, simProxyWrap);
			}
		}

		private void OnDisable()
		{
			if(originalPrimarySkinWrap != null){ 
				originalPrimarySkinWrap.renderSuspend = false;
				originalPrimarySkinWrap.enabled = true;
			}
			if(primaryWrap != null) {
				primaryWrap.renderSuspend = true;
				primaryWrap.enabled = false;
				this.SwapSkinWrap(primaryWrap, originalPrimarySkinWrap);
			}
			if(simProxyWrap != null) {
				this.SwapSkinWrap(simProxyWrap, originalProxySkinWrap);
			}
		}
	}
}